import numpy as np
import pandas as pd
import pyarrow.dataset as ds
import pyarrow.parquet as pq
from hats.io import file_io, paths
from hats.pixel_math.healpix_pixel import HealpixPixel
from hats.pixel_math.spatial_index import SPATIAL_INDEX_COLUMN, spatial_index_to_healpix

from hats_import.margin_cache.margin_cache_resume_plan import MarginCachePlan
from hats_import.pipeline_resume_plan import get_pixel_cache_directory, print_task_failure


# pylint: disable=too-many-arguments,disable=too-many-locals
def map_pixel_shards(
    partition_file,
    source_pixel,
    mapping_key,
    original_catalog_metadata,
    margin_pair_file,
    output_path,
    margin_order,
):
    """Creates margin cache shards from a source partition file."""
    try:
        schema = file_io.read_parquet_metadata(original_catalog_metadata).schema.to_arrow_schema()
        data = pq.read_table(
            partition_file.path, filesystem=partition_file.fs, schema=schema
        ).combine_chunks()

        # Constrain the possible margin pairs, first by only those `margin_order` pixels
        # that **can** be contained in source pixel, then by `margin_order` pixels for rows
        # in source data
        margin_pairs = pd.read_csv(margin_pair_file)
        explosion_factor = 4 ** int(margin_order - source_pixel.order)
        margin_pixel_range_start = source_pixel.pixel * explosion_factor
        margin_pixel_range_end = (source_pixel.pixel + 1) * explosion_factor
        margin_pairs = margin_pairs.query(
            f"margin_pixel >= {margin_pixel_range_start} and margin_pixel < {margin_pixel_range_end}"
        )

        margin_pixel_list = spatial_index_to_healpix(
            data[SPATIAL_INDEX_COLUMN].to_numpy(),
            target_order=margin_order,
        )
        margin_pixel_filter = pd.DataFrame(
            {"margin_pixel": margin_pixel_list, "filter_value": np.arange(0, len(margin_pixel_list))}
        ).merge(margin_pairs, on="margin_pixel")

        # For every possible output pixel, find the full margin_order pixel filter list,
        # perform the filter, and pass along to helper method to compute fine filter
        # and write out shard file.
        num_rows = 0
        for partition_key, data_filter in margin_pixel_filter.groupby(["partition_order", "partition_pixel"]):
            data_filter = np.unique(data_filter["filter_value"])
            filtered_data = data.take(data_filter)
            pixel = HealpixPixel(partition_key[0], partition_key[1])

            num_rows += _to_pixel_shard(
                filtered_data=filtered_data,
                pixel=pixel,
                output_path=output_path,
                source_pixel=source_pixel,
            )

        MarginCachePlan.mapping_key_done(output_path, mapping_key, num_rows)
    except Exception as exception:  # pylint: disable=broad-exception-caught
        print_task_failure(f"Failed MAPPING stage for pixel: {mapping_key}", exception)
        raise exception


# pylint: disable=too-many-arguments, unused-argument
def _to_pixel_shard(
    filtered_data,
    pixel,
    output_path,
    source_pixel,
):
    """Do boundary checking for the cached partition and then output remaining data."""
    margin_data = filtered_data

    num_rows = len(margin_data)
    if num_rows:
        # generate a file name for our margin shard, that uses both sets of Norder/Npix
        partition_dir = get_pixel_cache_directory(output_path, pixel)
        shard_dir = paths.pixel_directory(partition_dir, source_pixel.order, source_pixel.pixel)

        file_io.make_directory(shard_dir, exist_ok=True)

        shard_path = paths.pixel_catalog_file(partition_dir, source_pixel)

        margin_data = margin_data.sort_by(SPATIAL_INDEX_COLUMN)

        pq.write_table(margin_data, shard_path.path, filesystem=shard_path.fs)
    return num_rows


def reduce_margin_shards(
    intermediate_directory,
    reducing_key,
    output_path,
    partition_order,
    partition_pixel,
    delete_intermediate_parquet_files,
):
    """Reduce all partition pixel directories into a single file"""
    try:
        healpix_pixel = HealpixPixel(partition_order, partition_pixel)
        shard_dir = get_pixel_cache_directory(intermediate_directory, healpix_pixel)

        if file_io.does_file_or_directory_exist(shard_dir):
            margin_table = ds.dataset(shard_dir.path, filesystem=shard_dir.fs, format="parquet").to_table()

            if len(margin_table):
                margin_cache_dir = paths.pixel_directory(output_path, partition_order, partition_pixel)
                file_io.make_directory(margin_cache_dir, exist_ok=True)

                margin_cache_file_path = paths.pixel_catalog_file(output_path, healpix_pixel)
                pq.write_table(
                    margin_table, margin_cache_file_path.path, filesystem=margin_cache_file_path.fs
                )

                if delete_intermediate_parquet_files:
                    file_io.remove_directory(shard_dir)

        MarginCachePlan.reducing_key_done(intermediate_directory, reducing_key)
    except Exception as exception:  # pylint: disable=broad-exception-caught
        print_task_failure(f"Failed REDUCING stage for pixel: {reducing_key}", exception)
        raise exception



from astropy.io import ascii as ascii_reader

from hats_import.catalog.file_readers.input_reader import InputReader


class AstropyEcsvReader(InputReader):
    """Reads astropy ascii .ecsv files.

    Note that this is NOT a chunked reader. Use caution when reading
    large ECSV files with this reader.

    Attributes:
        kwargs: keyword arguments passed to astropy ascii reader.
            See https://docs.astropy.org/en/stable/api/astropy.io.ascii.read.html#astropy.io.ascii.read
    """

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def read(self, input_file, read_columns=None):
        self.regular_file_exists(input_file, **self.kwargs)
        if read_columns:
            self.kwargs["include_names"] = read_columns

        astropy_table = ascii_reader.read(input_file, format="ecsv", **self.kwargs)
        yield astropy_table.to_pandas()



import hats
import pandas as pd
import pyarrow.parquet as pq
import pytest
from hats import read_hats
from hats.io import paths

import hats_import.catalog.run_import as runner
from hats_import.catalog import ImportArguments
from hats_import.catalog.file_readers import ParquetPyarrowReader


def test_reimport_arguments(tmp_path, small_sky_object_catalog):
    args = ImportArguments.reimport_from_hats(
        small_sky_object_catalog, tmp_path, addl_hats_properties={"obs_regime": "Optical"}
    )
    catalog = hats.read_hats(small_sky_object_catalog)
    file_paths = [
        hats.io.pixel_catalog_file(catalog.catalog_base_dir, p) for p in catalog.get_healpix_pixels()
    ]
    assert args.catalog_type == catalog.catalog_info.catalog_type
    assert args.ra_column == catalog.catalog_info.ra_column
    assert args.dec_column == catalog.catalog_info.dec_column
    assert args.input_paths == file_paths
    assert isinstance(args.file_reader, ParquetPyarrowReader)
    assert args.output_artifact_name == catalog.catalog_name
    assert args.expected_total_rows == catalog.catalog_info.total_rows
    assert args.addl_hats_properties == catalog.catalog_info.extra_dict(by_alias=True) | {
        "hats_cols_default": catalog.catalog_info.default_columns,
        "hats_npix_suffix": catalog.catalog_info.npix_suffix,
        "obs_regime": "Optical",
    }
    assert args.use_healpix_29
    assert not args.add_healpix_29


def test_reimport_arguments_constant(tmp_path, small_sky_object_catalog):
    args = ImportArguments.reimport_from_hats(small_sky_object_catalog, tmp_path, constant_healpix_order=6)
    catalog = hats.read_hats(small_sky_object_catalog)
    file_paths = [
        hats.io.pixel_catalog_file(catalog.catalog_base_dir, p) for p in catalog.get_healpix_pixels()
    ]
    assert args.catalog_type == catalog.catalog_info.catalog_type
    assert args.ra_column == catalog.catalog_info.ra_column
    assert args.dec_column == catalog.catalog_info.dec_column
    assert args.input_paths == file_paths
    assert isinstance(args.file_reader, ParquetPyarrowReader)
    assert args.output_artifact_name == catalog.catalog_name
    assert args.expected_total_rows == catalog.catalog_info.total_rows
    assert args.addl_hats_properties == catalog.catalog_info.extra_dict(by_alias=True) | {
        "hats_cols_default": catalog.catalog_info.default_columns,
        "hats_npix_suffix": catalog.catalog_info.npix_suffix,
    }
    assert args.use_healpix_29
    assert not args.add_healpix_29


def test_reimport_arguments_extra_kwargs(tmp_path, small_sky_object_catalog):
    output_name = "small_sky_higher_order"
    pixel_thresh = 100
    args = ImportArguments.reimport_from_hats(
        small_sky_object_catalog,
        tmp_path,
        pixel_threshold=pixel_thresh,
        output_artifact_name=output_name,
        highest_healpix_order=2,
    )
    catalog = hats.read_hats(small_sky_object_catalog)
    file_paths = [
        hats.io.pixel_catalog_file(catalog.catalog_base_dir, p) for p in catalog.get_healpix_pixels()
    ]
    assert args.catalog_type == catalog.catalog_info.catalog_type
    assert args.ra_column == catalog.catalog_info.ra_column
    assert args.dec_column == catalog.catalog_info.dec_column
    assert args.input_paths == file_paths
    assert isinstance(args.file_reader, ParquetPyarrowReader)
    assert args.output_artifact_name == output_name
    assert args.pixel_threshold == pixel_thresh
    assert args.expected_total_rows == catalog.catalog_info.total_rows
    assert args.addl_hats_properties == catalog.catalog_info.extra_dict(by_alias=True) | {
        "hats_cols_default": catalog.catalog_info.default_columns,
        "hats_npix_suffix": catalog.catalog_info.npix_suffix,
    }
    assert args.use_healpix_29
    assert not args.add_healpix_29


def test_reimport_arguments_empty_dir(tmp_path):
    wrong_input_path = tmp_path / "nonsense"
    with pytest.raises(FileNotFoundError):
        ImportArguments.reimport_from_hats(wrong_input_path, tmp_path)


def test_reimport_arguments_invalid_dir(wrong_files_and_rows_dir, tmp_path):
    with pytest.raises(ValueError):
        ImportArguments.reimport_from_hats(wrong_files_and_rows_dir, tmp_path)


def test_reimport_arguments_catalog_collection(test_data_dir, small_sky_object_catalog, tmp_path):
    wrong_input_path = test_data_dir / "small_sky_collection"
    args = ImportArguments.reimport_from_hats(wrong_input_path, tmp_path)

    catalog = hats.read_hats(small_sky_object_catalog)
    assert len(args.input_paths) == len(catalog.get_healpix_pixels())
    assert args.catalog_type == catalog.catalog_info.catalog_type
    assert args.ra_column == catalog.catalog_info.ra_column
    assert args.dec_column == catalog.catalog_info.dec_column


@pytest.mark.dask(timeout=10)
def test_run_reimport(
    dask_client,
    small_sky_object_catalog,
    tmp_path,
):
    output_name = "small_sky_higher_order"
    pixel_thresh = 100
    args = ImportArguments.reimport_from_hats(
        small_sky_object_catalog,
        tmp_path,
        pixel_threshold=pixel_thresh,
        output_artifact_name=output_name,
        highest_healpix_order=1,
        addl_hats_properties={"obs_regime": "Optical"},
    )

    runner.run(args, dask_client)

    old_cat = read_hats(small_sky_object_catalog)

    # Check that the catalog metadata file exists
    catalog = read_hats(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert catalog.catalog_info.ra_column == old_cat.catalog_info.ra_column
    assert catalog.catalog_info.dec_column == old_cat.catalog_info.dec_column
    assert catalog.catalog_info.total_rows == old_cat.catalog_info.total_rows
    assert len(old_cat.catalog_info.default_columns) > 0
    assert catalog.catalog_info.default_columns == old_cat.catalog_info.default_columns
    assert catalog.catalog_info.__pydantic_extra__["obs_regime"] == "Optical"
    assert len(catalog.get_healpix_pixels()) == 4

    # Check that the schema is correct for leaf parquet and _metadata files
    original_common_md = paths.get_common_metadata_pointer(old_cat.catalog_base_dir)
    expected_parquet_schema = pq.read_metadata(original_common_md).schema.to_arrow_schema()
    new_schema = paths.get_common_metadata_pointer(catalog.catalog_base_dir)
    schema = pq.read_metadata(new_schema).schema.to_arrow_schema()
    assert schema.equals(expected_parquet_schema)
    schema = pq.read_metadata(args.catalog_path / "dataset" / "_metadata").schema.to_arrow_schema()
    assert schema.equals(expected_parquet_schema)
    output_file = args.catalog_path / "dataset" / "Norder=1" / "Dir=0" / "Npix=44.parquet"
    schema = pq.read_metadata(output_file).schema.to_arrow_schema()
    assert schema.equals(expected_parquet_schema)

    # Check that, when re-loaded as a pandas dataframe, the appropriate numeric types are used.
    data_frame = pd.read_parquet(output_file, engine="pyarrow")
    expected_dtypes = expected_parquet_schema.empty_table().to_pandas().dtypes
    assert data_frame.dtypes.equals(expected_dtypes)



"""Tests of margin cache generation arguments"""

import pytest
from hats.pixel_math.healpix_pixel import HealpixPixel

from hats_import.margin_cache.margin_cache_arguments import MarginCacheArguments


def test_empty_required(tmp_path):
    """*Most* required arguments are provided."""
    ## Input catalog path is missing
    with pytest.raises(ValueError, match="input_catalog_path"):
        MarginCacheArguments(
            margin_threshold=5.0,
            output_path=tmp_path,
            output_artifact_name="catalog_cache",
        )

    ## Input catalog path is bad
    with pytest.raises(ValueError, match="input_catalog_path"):
        MarginCacheArguments(
            input_catalog_path="/foo",
            margin_threshold=5.0,
            output_path=tmp_path,
            output_artifact_name="catalog_cache",
        )


def test_margin_order_dynamic(small_sky_source_catalog, tmp_path):
    """Ensure we can dynamically set the margin_order"""
    args = MarginCacheArguments(
        margin_threshold=5.0,
        input_catalog_path=small_sky_source_catalog,
        output_path=tmp_path,
        output_artifact_name="catalog_cache",
    )

    assert args.margin_order == 14


def test_margin_order_static(small_sky_source_catalog, tmp_path):
    """Ensure we can manually set the margin_order"""
    args = MarginCacheArguments(
        margin_threshold=5.0,
        input_catalog_path=small_sky_source_catalog,
        output_path=tmp_path,
        output_artifact_name="catalog_cache",
        margin_order=4,
    )

    assert args.margin_order == 4


def test_margin_order_invalid_small_margin_order(small_sky_source_catalog, tmp_path):
    """Ensure we raise an exception when margin_order is invalid"""
    with pytest.raises(ValueError, match="margin_order"):
        MarginCacheArguments(
            margin_threshold=5.0,
            input_catalog_path=small_sky_source_catalog,
            output_path=tmp_path,
            output_artifact_name="catalog_cache",
            margin_order=2,
        )


def test_margin_threshold_invalid_large_margin_threshold(small_sky_source_catalog, tmp_path):
    """Ensure we give a warning when margin_threshold is greater than margin_order resolution"""

    with pytest.raises(ValueError, match="margin_threshold"):
        MarginCacheArguments(
            margin_threshold=360.0,
            input_catalog_path=small_sky_source_catalog,
            output_path=tmp_path,
            output_artifact_name="catalog_cache",
            margin_order=16,
        )


def test_debug_filter_pixel_list(small_sky_source_catalog, tmp_path):
    """Ensure we can generate catalog with a filtereed list of pixels, and
    that we raise an exception when the filter results in an empty catalog."""
    args = MarginCacheArguments(
        margin_threshold=5.0,
        input_catalog_path=small_sky_source_catalog,
        output_path=tmp_path,
        output_artifact_name="catalog_cache",
        margin_order=4,
        debug_filter_pixel_list=[HealpixPixel(0, 11)],
    )

    assert len(args.catalog.get_healpix_pixels()) == 13

    args = MarginCacheArguments(
        margin_threshold=5.0,
        input_catalog_path=small_sky_source_catalog,
        output_path=tmp_path,
        output_artifact_name="catalog_cache",
        margin_order=4,
        debug_filter_pixel_list=[HealpixPixel(1, 44)],
    )

    assert len(args.catalog.get_healpix_pixels()) == 4

    with pytest.raises(ValueError, match="debug_filter_pixel_list"):
        MarginCacheArguments(
            margin_threshold=5.0,
            input_catalog_path=small_sky_source_catalog,
            output_path=tmp_path,
            output_artifact_name="catalog_cache",
            margin_order=4,
            debug_filter_pixel_list=[HealpixPixel(0, 5)],
        )


def test_to_table_properties(small_sky_source_catalog, tmp_path):
    """Verify creation of catalog info for margin cache to be created."""
    args = MarginCacheArguments(
        margin_threshold=5.0,
        input_catalog_path=small_sky_source_catalog,
        output_path=tmp_path,
        output_artifact_name="catalog_cache",
        margin_order=4,
    )
    catalog_info = args.to_table_properties(total_rows=10, highest_order=4, moc_sky_fraction=22 / 7)
    assert catalog_info.catalog_name == args.output_artifact_name
    assert catalog_info.total_rows == 10
    assert catalog_info.ra_column == "source_ra"
    assert catalog_info.dec_column == "source_dec"


def test_no_margin_cache_overwrite(small_sky_object_catalog):
    """Runner should refuse to generate margin cache which overwrites valid catalog"""
    catalog_dir = small_sky_object_catalog.parent
    catalog_name = small_sky_object_catalog.name
    with pytest.raises(ValueError, match="already contains a valid catalog"):
        MarginCacheArguments(
            input_catalog_path=small_sky_object_catalog,
            output_path=catalog_dir,
            margin_threshold=10.0,
            output_artifact_name=catalog_name,
        )



from __future__ import annotations

from dataclasses import dataclass, field

import hats.pixel_math.healpix_shim as hp
from hats import read_hats
from hats.catalog.dataset.collection_properties import CollectionProperties
from hats.io import file_io
from hats.io.validation import is_valid_catalog
from upath import UPath

from hats_import.catalog import ImportArguments
from hats_import.index.arguments import IndexArguments
from hats_import.margin_cache.margin_cache_arguments import MarginCacheArguments
from hats_import.runtime_arguments import RuntimeArguments


@dataclass
class CollectionArguments(RuntimeArguments):
    """Container class for building arguments for importing catalog as a collection."""

    new_catalog_path: UPath | None = None
    """Constructed path for the new catalog, relative to the collection."""
    new_catalog_name: str | None = None
    """Name for the new catalog that will be created. Taken from the ``output_artifact_name`` of
    catalog kwargs. Used to construct names of the supplemental catalogs, if their own
    ``output_artifact_name`` isn't specified."""
    catalog_args: ImportArguments | None = None
    """Constructed arguments for catalog import."""
    margin_kwargs: list[dict] = field(default_factory=list)
    """List of all argument dictionaries passed to this builder, for creating margins."""
    default_margin_name: str | None = None
    index_kwargs: list[dict] = field(default_factory=list)
    """List of all argument dictionaries passed to this builder, for creating indexes."""
    margin_args: list[dict] = field(default_factory=list)
    """Constructed arguments for creating margins."""
    index_args: list[dict] = field(default_factory=list)
    """Constructed arguments for creating indexes."""

    def catalog(self, **kwargs):
        """Set the primary catalog for the collection.

        NB: This should only be called EXACTLY ONCE per catalog collection.
        If building a new catalog from scratch, these should be all the arguments necessary
        for import. If using an existing catalog, you should provide the ``output_artifact_name``
        so the primary catalog subdirectory can be located and validated.
        """
        if self.new_catalog_path is not None:
            raise ValueError("Must call catalog method exactly once.")
        useful_kwargs = self._get_subarg_dict()
        useful_kwargs.update(kwargs)
        if "output_artifact_name" not in kwargs:
            useful_kwargs["output_artifact_name"] = self.output_artifact_name

        ## Test for an existing catalog at the indicated path.
        if "catalog_path" in kwargs:
            new_catalog_path = file_io.get_upath(kwargs["catalog_path"])
        else:
            new_catalog_path = (
                file_io.get_upath(self.output_path)
                / self.output_artifact_name
                / useful_kwargs["output_artifact_name"]
            )
        if is_valid_catalog(new_catalog_path):
            ## There is already a valid catalog (either from resume or pre-existing).
            ## Leave it alone and write the remainder of the collection contents.
            catalog = read_hats(new_catalog_path)
            self.new_catalog_path = catalog.catalog_path
            self.new_catalog_name = catalog.catalog_name
            return self

        self.catalog_args = ImportArguments(**useful_kwargs)
        self.new_catalog_path = self.catalog_args.catalog_path
        self.new_catalog_name = self.catalog_args.output_artifact_name

        return self

    def get_catalog_args(self):
        """Retrieve the catalog arguments, if a catalog must be created or resumed."""
        if self.new_catalog_path is None:
            raise ValueError("Must add catalog arguments before fetching catalog arguments")

        return self.catalog_args

    def add_margin(self, is_default=False, **kwargs):
        """Add arguments for a margin catalog.

        NB: This can be called 0, 1, or many times for a single collection.
        This method will only stash the provided arguments for later use,
        as the arguments cannot be validated until the catalog exists on disk.

        Args:
            is_default(bool): If True, this margin will be set as the default on the
                catalog collection's properties.
            kwargs (dict): arguments passed to ``MarginCacheArguments`` constructor.
        """
        if self.new_catalog_path is None:
            raise ValueError("Must add catalog arguments before adding margin arguments")

        if "margin_threshold" not in kwargs and "margin_order" not in kwargs:
            raise ValueError("Margin threshold required, either in arcseconds or HEALPix order.")

        margin_threshold = kwargs["margin_threshold"] or hp.order2mindist(kwargs["margin_order"]) * 60

        useful_kwargs = self._get_subarg_dict()
        useful_kwargs.update(kwargs)

        if "output_artifact_name" not in kwargs:
            margin_suffix = _pretty_print_angle(margin_threshold)
            useful_kwargs["output_artifact_name"] = f"{self.new_catalog_name}_{margin_suffix}"
        if is_default:
            default_name = useful_kwargs["output_artifact_name"]
            if self.default_margin_name:
                raise ValueError(
                    "Only one margin catalog may be the default "
                    f"({self.default_margin_name} already set, trying to add {default_name})"
                )
            self.default_margin_name = default_name

        if "input_catalog_path" not in kwargs:
            useful_kwargs["input_catalog_path"] = self.new_catalog_path

        self.margin_kwargs.append(useful_kwargs)

        return self

    def get_margin_args(self):
        """Construct and return the margin argument objects, validating the inputs."""
        if self.new_catalog_path is None:
            raise ValueError("Must add catalog arguments before fetching margin arguments")
        if not self.margin_args:
            self.margin_args = []
            for margin_kwargs in self.margin_kwargs:
                self.margin_args.append(MarginCacheArguments(**margin_kwargs))
        return self.margin_args

    def add_index(self, **kwargs):
        """Add arguments for an index catalog.

        NB: This can be called 0, 1, or many times for a single collection.
        This method will only stash the provided arguments for later use,
        as the arguments cannot be validated until the catalog exists on disk."""
        if self.new_catalog_path is None:
            raise ValueError("Must add catalog arguments before adding index arguments")

        if "indexing_column" not in kwargs:
            raise ValueError("indexing_column is required")

        useful_kwargs = self._get_subarg_dict()
        useful_kwargs.update(kwargs)

        if "output_artifact_name" not in kwargs:
            useful_kwargs["output_artifact_name"] = f"{self.new_catalog_name}_{kwargs['indexing_column']}"
        if "input_catalog_path" not in kwargs:
            useful_kwargs["input_catalog_path"] = self.new_catalog_path

        self.index_kwargs.append(useful_kwargs)

        return self

    def get_index_args(self):
        """Construct and return the index argument objects, validating the inputs."""
        if self.new_catalog_path is None:
            raise ValueError("Must add catalog arguments before fetching index arguments")
        if not self.index_args:
            self.index_args = []
            for index_kwargs in self.index_kwargs:
                self.index_args.append(IndexArguments(**index_kwargs))
        return self.index_args

    def _get_subarg_dict(self):
        """Get a subset of this object's arguments as a dictionary.

        Not all of the original arguments are useful, and ALL of them
        can potentially be overriden by the user's ``kwargs`` on the method.

        Note that we will not copy over arguments related to dask runtime, as
        the client will be created once at the beginning of collection creation
        and used throughout.
        """
        useful_kwargs = {
            "output_path": self.catalog_path,
            "addl_hats_properties": self.addl_hats_properties,
            "tmp_dir": self.tmp_path,
            "resume": self.resume,
            "progress_bar": self.progress_bar,
            "simple_progress_bar": self.simple_progress_bar,
            "delete_intermediate_parquet_files": self.delete_intermediate_parquet_files,
            "delete_resume_log_files": self.delete_resume_log_files,
        }
        return useful_kwargs

    def to_collection_properties(self):
        """Collection-specific dataset info."""
        if self.new_catalog_path is None:
            raise ValueError("Must add catalog arguments before collection properties")
        info = {"obs_collection": self.output_artifact_name}

        def _maybe_relative(artifact_path, collection_path):
            if artifact_path.is_relative_to(collection_path):
                return str(artifact_path.relative_to(collection_path))
            return str(artifact_path)

        info["hats_primary_table_url"] = _maybe_relative(self.new_catalog_path, self.catalog_path)

        margin_paths = [
            _maybe_relative(args.catalog_path, self.catalog_path) for args in self.get_margin_args()
        ]
        if margin_paths:
            info["all_margins"] = margin_paths
        if self.default_margin_name:
            info["default_margin"] = self.default_margin_name

        index_paths = {
            args.indexing_column: _maybe_relative(args.catalog_path, self.catalog_path)
            for args in self.get_index_args()
        }
        if index_paths:
            info["all_indexes"] = index_paths

        info = info | self.extra_property_dict()
        properties = CollectionProperties(**info)
        return properties


def _pretty_print_angle(arc_seconds):
    if arc_seconds >= 3600:
        return f"{int(arc_seconds / 3600)}deg"
    if arc_seconds >= 60:
        return f"{int(arc_seconds/60)}arcmin"
    if arc_seconds >= 1:
        return f"{int(arc_seconds)}arcs"
    return f"{int(arc_seconds * 1000)}msec"



"""All modules for hats-import package"""

from ._version import __version__
from .catalog import ImportArguments
from .collection import CollectionArguments
from .index import IndexArguments
from .margin_cache.margin_cache_arguments import MarginCacheArguments
from .pipeline import pipeline, pipeline_with_client
from .runtime_arguments import RuntimeArguments
from .verification.arguments import VerificationArguments



"""Utility to hold a pipeline's execution and resume plan."""

from __future__ import annotations

import re
from dataclasses import dataclass
from pathlib import Path

from dask.distributed import as_completed, get_worker
from dask.distributed import print as dask_print
from hats.io import file_io
from hats.pixel_math.healpix_pixel import HealpixPixel
from tqdm.auto import tqdm as auto_tqdm
from tqdm.std import tqdm as std_tqdm
from upath import UPath


@dataclass
class PipelineResumePlan:
    """Container class for holding the state of pipeline plan."""

    tmp_path: UPath
    """path for any intermediate files"""
    tmp_base_path: str | Path | UPath | None = None
    """temporary base directory: either `tmp_dir` or `dask_dir`, if those were provided by the user"""
    resume: bool = True
    """if there are existing intermediate resume files, should we
    read those and continue to run pipeline where we left off"""
    progress_bar: bool = True
    """if true, a tqdm progress bar will be displayed for user
    feedback of planning progress"""
    simple_progress_bar: bool = False
    """if displaying a progress bar, use a text-only simple progress
    bar instead of widget. this can be useful in some environments when running
    in a notebook where ipywidgets cannot be used (see `progress_bar` argument)"""
    delete_resume_log_files: bool = True
    """should we delete task-level done files once each stage is complete?
    if False, we will keep all sub-histograms from the mapping stage, and all
    done marker files at the end of the pipeline."""
    delete_intermediate_parquet_files: bool = True
    """should we delete any smaller intermediate parquet files that may be
    generated by the pipeline?"""

    ORIGINAL_INPUT_PATHS = "input_paths.txt"

    def safe_to_resume(self):
        """Check that we are ok to resume an in-progress pipeline, if one exists.

        Raises:
            ValueError: if the tmp_path already exists and contains some files.
        """
        if file_io.directory_has_contents(self.tmp_path):
            if not self.resume:
                file_io.remove_directory(self.tmp_path, ignore_errors=True)
            else:
                print(f"tmp_path ({self.tmp_path}) contains intermediate files; resuming prior progress.")
        file_io.make_directory(self.tmp_path, exist_ok=True)

    def done_file_exists(self, stage_name):
        """Is there a file at a given path?
        For a done file, the existence of the file is the only signal needed to indicate
        a pipeline segment is complete.

        Args:
            stage_name(str): name of the stage (e.g. mapping, reducing)
        Returns:
            boolean, True if the done file exists at tmp_path. False otherwise.
        """
        return file_io.does_file_or_directory_exist(
            file_io.append_paths_to_pointer(self.tmp_path, f"{stage_name}_done")
        )

    def touch_stage_done_file(self, stage_name):
        """Touch (create) a done file for a whole pipeline stage.
        For a done file, the existence of the file is the only signal needed to indicate
        a pipeline segment is complete.

        Args:
            stage_name(str): name of the stage (e.g. mapping, reducing)
        """
        Path(file_io.append_paths_to_pointer(self.tmp_path, f"{stage_name}_done")).touch()

    @classmethod
    def touch_key_done_file(cls, tmp_path, stage_name, key):
        """Touch (create) a done file for a single key, within a pipeline stage.

        Args:
            tmp_path: where to write pipeline intermediate files
            stage_name(str): name of the stage (e.g. mapping, reducing)
            key (str): unique string for each task (e.g. "map_57")
        """
        Path(file_io.append_paths_to_pointer(tmp_path, stage_name, f"{key}_done")).touch()

    @classmethod
    def write_marker_file(cls, tmp_path, stage_name, key, value):
        """Create a marker file for a single key, within a pipeline stage.

        Similar to a "done" file, but contains some value inside the file to be read later.

        Args:
            tmp_path: where to write pipeline intermediate files
            stage_name(str): name of the stage (e.g. mapping, reducing)
            key (str): unique string for each task (e.g. "map_57")
            value (str): value for the marker.
        """
        file_io.write_string_to_file(
            file_io.append_paths_to_pointer(tmp_path, stage_name, f"{key}_done"), value
        )

    def read_markers(self, stage_name: str) -> dict[str, list[str]]:
        """Inspect the stage's directory of marker files, fetching the key value pairs
        from marker file names and contents.

        Args:
            stage_name(str): name of the stage (e.g. mapping, reducing)
        Return:
            list[str] - all keys found in done directory
        """
        prefix = file_io.append_paths_to_pointer(self.tmp_path, stage_name)
        result = {}
        result_files = file_io.find_files_matching_path(prefix, "*_done")
        done_file_pattern = re.compile(r"(.*)_done")
        for file_path in result_files:
            match = done_file_pattern.match(str(file_path.name))
            if not match:
                raise ValueError(f"Unexpected file found: {file_path.name}")
            key = match.group(1)
            result[key] = file_io.load_text_file(file_path)
        return result

    def read_done_pixels(self, stage_name):
        """Inspect the stage's directory of done files, fetching the pixel keys from done file names.

        Args:
            stage_name(str): name of the stage (e.g. mapping, reducing)
        Return:
            List[HealpixPixel] - all pixel keys found in done directory
        """
        prefix = file_io.append_paths_to_pointer(self.tmp_path, stage_name)
        done_file_pattern = re.compile(r"(\d+)_(\d+)_done")
        pixel_tuples = [done_file_pattern.match(path.name).group(1, 2) for path in prefix.glob("*_done")]
        return [HealpixPixel(int(match[0]), int(match[1])) for match in pixel_tuples]

    def clean_resume_files(self):
        """Remove the intermediate directory created in execution if the user decided
        to erase all the stage log files as well as intermediate parquet."""
        if self.delete_resume_log_files and self.delete_intermediate_parquet_files:
            # Use the temporary directory base path if the user provided it, or
            # delete the intermediate directory from inside the output directory
            path = self.tmp_base_path if self.tmp_base_path is not None else self.tmp_path
            file_io.remove_directory(path, ignore_errors=True)

    def wait_for_futures(self, futures, stage_name, fail_fast=False):
        """Wait for collected futures to complete.

        As each future completes, check the returned status.

        Args:
            futures(List[future]): collected futures
            stage_name(str): name of the stage (e.g. mapping, reducing)
            fail_fast (bool): if True, we will re-raise the first exception
                encountered and NOT continue. this may lead to unexpected
                behavior for in-progress tasks.
        Raises:
            RuntimeError: if any future returns an error status.
        """
        some_error = False
        for future in self.print_progress(as_completed(futures), stage_name=stage_name, total=len(futures)):
            if future.status == "error":
                some_error = True
                if fail_fast:
                    raise future.exception()

        if some_error:
            raise RuntimeError(f"Some {stage_name} stages failed. See above Exceptions.")

    def print_progress(self, iterable=None, total=None, stage_name=None):
        """Create a progress bar that will provide user with task feedback.

        This is a thin wrapper around the static ``print_progress`` method that uses
        member variables for the caller's convenience.

        Args:
            iterable (iterable): Optional. provides iterations to progress updates.
            total (int): Optional. Expected iterations.
            stage_name (str): name of the stage (e.g. mapping, reducing). this will
                be further formatted with ``get_formatted_stage_name``, so the caller
                doesn't need to worry about that.
        """
        return print_progress(
            iterable=iterable,
            total=total,
            stage_name=stage_name,
            use_progress_bar=self.progress_bar,
            simple_progress_bar=self.simple_progress_bar,
        )

    def check_original_input_paths(self, input_paths):
        """Validate that we're operating on the same file set as the original pipeline,
        or save the inputs as the originals if not found.

        Args:
            input_paths (list[str]): input paths that will be processed by a pipeline.

        Raises:
            ValueError: if the retrieved file set differs from `input_paths`.
        """
        if not input_paths:
            return []
        expected_input_paths = set(input_paths)
        expected_input_paths = [str(p) for p in input_paths]
        expected_input_paths.sort()

        original_input_paths = []

        log_file_path = file_io.append_paths_to_pointer(self.tmp_path, self.ORIGINAL_INPUT_PATHS)
        try:
            with open(log_file_path, "r", encoding="utf-8") as file_handle:
                contents = file_handle.readlines()
            contents = [path.strip() for path in contents]
            original_input_paths = list(set(contents))
            original_input_paths.sort()
        except FileNotFoundError:
            pass

        if len(original_input_paths) == 0:
            with open(log_file_path, "w", encoding="utf-8") as file_handle:
                for path in input_paths:
                    file_handle.write(f"{path}\n")
        else:
            if original_input_paths != expected_input_paths:
                raise ValueError("Different file set from resumed pipeline execution.")

        return input_paths


def get_pixel_cache_directory(cache_path, pixel: HealpixPixel):
    """Create a path for intermediate pixel data.

    You can use this over the paths.get_pixel_directory method, as it will include the pixel
    number in the path. Further, it will just *look* different from a real hats
    path, so it's clearer that it's a temporary directory::

        {cache_path}/order_{order}/dir_{dir}/pixel_{pixel}/

    Args:
        cache_path (str): root path to cache
        pixel (HealpixPixel): pixel partition data
    """
    return file_io.append_paths_to_pointer(
        cache_path, f"order_{pixel.order}", f"dir_{pixel.dir}", f"pixel_{pixel.pixel}"
    )


def print_task_failure(custom_message, exception):
    """Use dask's distributed print feature to print the exception message to the task's logs
    and to the controller job's logs.

    Optionally print the worker address if a worker is found.

    Args:
        custom_message (str): some custom message for the task that might help with debugging
        exception (Exception): error raised in execution of the task.
    """
    dask_print(custom_message)
    try:
        dask_print("  worker address:", get_worker().address)
    except Exception:  # pylint: disable=broad-exception-caught
        pass
    dask_print(exception)


def get_formatted_stage_name(stage_name) -> str:
    """Create a stage name of consistent minimum length. Ensures that the tqdm
    progress bars can line up nicely when multiple stages must run.

    Args:
        stage_name (str): name of the stage (e.g. mapping, reducing)
    """
    if stage_name is None or len(stage_name) == 0:
        stage_name = "progress"

    return f"{stage_name.capitalize(): <10}"


def print_progress(
    iterable=None, total=None, stage_name=None, use_progress_bar=True, simple_progress_bar=False
):
    """Create a progress bar that will provide user with task feedback.

    Args:
        iterable (iterable): Optional. provides iterations to progress updates.
        total (int): Optional. Expected iterations.
        stage_name (str): name of the stage (e.g. mapping, reducing). this will
            be further formatted with ``get_formatted_stage_name``, so the caller
            doesn't need to worry about that.
        use_progress_bar (bool): should we display any progress. typically False
            when no stdout is expected.
        simple_progress_bar (bool): if displaying a progress bar, use a text-only
            simple progress bar instead of widget. this can be useful when running
            in a particular notebook where ipywidgets cannot be used
            (only used when ``use_progress_bar`` is True)
    """
    if simple_progress_bar:
        return std_tqdm(
            iterable,
            desc=get_formatted_stage_name(stage_name),
            total=total,
            disable=not use_progress_bar,
        )

    return auto_tqdm(
        iterable,
        desc=get_formatted_stage_name(stage_name),
        total=total,
        disable=not use_progress_bar,
    )



from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path

import hats.pixel_math.healpix_shim as hp
from hats import read_hats
from hats.catalog import TableProperties
from hats.io.validation import is_valid_catalog
from hats.pixel_math.healpix_pixel import HealpixPixel
from upath import UPath

from hats_import.runtime_arguments import RuntimeArguments


@dataclass
class MarginCacheArguments(RuntimeArguments):
    """Container for margin cache generation arguments"""

    margin_threshold: float = 5.0
    """the size of the margin cache boundary, given in arcseconds. setting the
    `margin_threshold` to be greater than the resolution of a `margin_order`
    healpixel will result in a warning, as this may lead to data loss."""
    margin_order: int = -1
    """the order of healpixels that will be used to constrain the margin data before
    doing more precise boundary checking. this value must be greater than the highest
    order of healpix partitioning in the source catalog. if `margin_order` is left
    default or set to -1, then the `margin_order` will be set dynamically to the
    highest partition order plus 1."""
    fine_filtering: bool = False
    """should we perform the precise boundary checking? if false, some results may be
    greater than `margin_threshold` away from the border (but within `margin_order`)."""

    input_catalog_path: str | Path | UPath | None = None
    """the path to the hats-formatted input catalog."""
    debug_filter_pixel_list: list[HealpixPixel] = field(default_factory=list)
    """debug setting. if provided, we will first filter the catalog to the pixels
    provided. this can be useful for creating a margin over a subset of a catalog."""

    def __post_init__(self):
        self._check_arguments()

    def _check_arguments(self):
        super()._check_arguments()
        if not self.input_catalog_path:
            raise ValueError("input_catalog_path is required")
        if not is_valid_catalog(self.input_catalog_path):
            raise ValueError("input_catalog_path not a valid catalog")

        self.catalog = read_hats(self.input_catalog_path)
        if len(self.debug_filter_pixel_list) > 0:
            self.catalog = self.catalog.filter_from_pixel_list(self.debug_filter_pixel_list)
            if len(self.catalog.get_healpix_pixels()) == 0:
                raise ValueError("debug_filter_pixel_list has created empty catalog")

        if self.fine_filtering:
            raise NotImplementedError("Fine filtering temporarily removed.")

        highest_order = int(self.catalog.partition_info.get_highest_order())

        if self.margin_order < 0:
            self.margin_order = hp.margin2order(margin_thr_arcmin=self.margin_threshold / 60.0)

        if self.margin_order < highest_order + 1:
            raise ValueError(
                "margin_order must be of a higher order than the highest order catalog partition pixel."
            )

        margin_pixel_mindist = hp.order2mindist(self.margin_order)
        if margin_pixel_mindist * 60.0 < self.margin_threshold:
            raise ValueError("margin pixels must be larger than margin_threshold")

    def to_table_properties(
        self, total_rows: int, highest_order: int, moc_sky_fraction: float
    ) -> TableProperties:
        """Catalog-type-specific dataset info."""
        info = {
            "catalog_name": self.output_artifact_name,
            "total_rows": total_rows,
            "catalog_type": "margin",
            "ra_column": self.catalog.catalog_info.ra_column,
            "dec_column": self.catalog.catalog_info.dec_column,
            "primary_catalog": str(self.input_catalog_path),
            "margin_threshold": self.margin_threshold,
            "hats_order": highest_order,
            "moc_sky_fraction": f"{moc_sky_fraction:0.5f}",
        } | self.extra_property_dict()
        return TableProperties(**info)



"""Functional tests for catalog import"""

import os
import shutil
from importlib.metadata import version
from pathlib import Path

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import pytest
from hats import read_hats
from hats.pixel_math.sparse_histogram import SparseHistogram
from pyarrow.parquet import ParquetFile

import hats_import.catalog.run_import as runner
from hats_import.catalog.arguments import ImportArguments
from hats_import.catalog.file_readers import CsvReader
from hats_import.catalog.resume_plan import ResumePlan


def test_empty_args():
    """Runner should fail with empty arguments"""
    with pytest.raises(ValueError, match="args is required"):
        runner.run(None, None)


def test_bad_args():
    """Runner should fail with mis-typed arguments"""
    args = {"output_artifact_name": "bad_arg_type"}
    with pytest.raises(ValueError, match="ImportArguments"):
        runner.run(args, None)


@pytest.mark.dask
def test_resume_dask_runner(
    dask_client,
    small_sky_parts_dir,
    resume_dir,
    tmp_path,
    assert_parquet_file_ids,
):
    """Test execution in the presence of some resume files."""
    ## First, copy over our intermediate files.
    ## This prevents overwriting source-controlled resume files.
    intermediate_dir = tmp_path / "resume_catalog" / "intermediate"
    shutil.copytree(resume_dir / "intermediate", intermediate_dir)
    ## Now set up our resume files to match previous work.
    resume_tmp = tmp_path / "tmp" / "resume_catalog"
    ResumePlan(tmp_path=resume_tmp, progress_bar=False)
    histogram = SparseHistogram([11], [131], 0)
    empty = SparseHistogram([], [], 0)
    for file_index in range(0, 5):
        ResumePlan.touch_key_done_file(resume_tmp, ResumePlan.SPLITTING_STAGE, f"split_{file_index}")
        histogram_file = ResumePlan.partial_histogram_file(
            tmp_path=resume_tmp, mapping_key=f"map_{file_index}"
        )
        if file_index == 0:
            histogram.to_file(histogram_file)
        else:
            empty.to_file(histogram_file)

    ResumePlan.touch_key_done_file(resume_tmp, ResumePlan.REDUCING_STAGE, "0_11")

    shutil.copytree(resume_dir / "Norder=0", tmp_path / "resume_catalog" / "dataset" / "Norder=0")

    args = ImportArguments(
        output_artifact_name="resume_catalog",
        input_path=small_sky_parts_dir,
        file_reader="csv",
        output_path=tmp_path,
        dask_tmp=tmp_path,
        tmp_dir=tmp_path,
        resume_tmp=tmp_path / "tmp",
        highest_healpix_order=0,
        pixel_threshold=1000,
        progress_bar=False,
    )

    runner.run(args, dask_client)

    # Check that the catalog metadata file exists
    catalog = read_hats(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert catalog.catalog_info.ra_column == "ra"
    assert catalog.catalog_info.dec_column == "dec"
    assert catalog.catalog_info.total_rows == 131
    assert len(catalog.get_healpix_pixels()) == 1

    # Check that the catalog parquet file exists and contains correct object IDs
    output_file = Path(args.catalog_path) / "dataset" / "Norder=0" / "Dir=0" / "Npix=11.parquet"

    expected_ids = [*range(700, 831)]
    assert_parquet_file_ids(output_file, "id", expected_ids)

    ## Re-running the pipeline with fully done intermediate files
    ## should result in no changes to output files.
    shutil.copytree(resume_dir / "intermediate", resume_tmp)

    args = ImportArguments(
        output_artifact_name="resume",
        input_path=small_sky_parts_dir,
        file_reader="csv",
        output_path=tmp_path,
        dask_tmp=tmp_path,
        tmp_dir=tmp_path,
        highest_healpix_order=0,
        pixel_threshold=1000,
        progress_bar=False,
    )

    runner.run(args, dask_client)

    catalog = read_hats(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert catalog.catalog_info.ra_column == "ra"
    assert catalog.catalog_info.dec_column == "dec"
    assert catalog.catalog_info.total_rows == 131
    assert len(catalog.get_healpix_pixels()) == 1
    assert_parquet_file_ids(output_file, "id", expected_ids)


@pytest.mark.dask
def test_resume_dask_runner_diff_pixel_order(
    dask_client,
    small_sky_parts_dir,
    resume_dir,
    tmp_path,
    assert_parquet_file_ids,
):
    """Test execution in the presence of histogram files that are not compatible
    with the current HEALPix order."""
    ## First, copy over our intermediate files.
    ## This prevents overwriting source-controlled resume files.
    intermediate_dir = tmp_path / "resume_catalog" / "intermediate"
    shutil.copytree(resume_dir / "intermediate", intermediate_dir)

    ## Now set up our resume files to match previous work.
    resume_tmp = tmp_path / "tmp" / "resume_catalog"
    ResumePlan(tmp_path=resume_tmp, progress_bar=False)
    SparseHistogram([11], [131], 0).to_dense_file(resume_tmp / "mapping_histogram.npz")
    for file_index in range(0, 5):
        ResumePlan.touch_key_done_file(resume_tmp, ResumePlan.SPLITTING_STAGE, f"split_{file_index}")

    shutil.copytree(resume_dir / "Norder=0", tmp_path / "resume_catalog" / "Norder=0")

    with pytest.raises(ValueError, match="incompatible with the highest healpix order"):
        args = ImportArguments(
            output_artifact_name="resume_catalog",
            input_path=small_sky_parts_dir,
            file_reader="csv",
            output_path=tmp_path,
            dask_tmp=tmp_path,
            tmp_dir=tmp_path,
            resume_tmp=tmp_path / "tmp",
            constant_healpix_order=1,
            pixel_threshold=1000,
            progress_bar=False,
        )
        runner.run(args, dask_client)

    # Running with resume set to "False" will start the pipeline from scratch
    args = ImportArguments(
        output_artifact_name="resume_catalog",
        input_path=small_sky_parts_dir,
        file_reader="csv",
        output_path=tmp_path,
        dask_tmp=tmp_path,
        tmp_dir=tmp_path,
        resume_tmp=tmp_path / "tmp",
        constant_healpix_order=1,
        pixel_threshold=1000,
        progress_bar=False,
        resume=False,
    )
    runner.run(args, dask_client)

    # Check that the catalog metadata file exists
    catalog = read_hats(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert catalog.catalog_info.ra_column == "ra"
    assert catalog.catalog_info.dec_column == "dec"
    assert catalog.catalog_info.total_rows == 131
    assert len(catalog.get_healpix_pixels()) == 4

    for n_pix in range(44, 48):
        output_filepath = args.catalog_path / "dataset" / "Norder=1" / "Dir=0" / f"Npix={n_pix}.parquet"
        expected_filepath = resume_dir / "Norder=1" / "Dir=0" / f"Npix={n_pix}.parquet"
        expected_file = pd.read_parquet(expected_filepath, engine="pyarrow")
        assert_parquet_file_ids(output_filepath, "id", expected_file["id"].to_numpy())


@pytest.mark.dask
def test_resume_dask_runner_histograms_diff_size(
    dask_client,
    small_sky_parts_dir,
    tmp_path,
):
    """Tests that the pipeline errors if the partial histograms have different sizes."""
    resume_tmp = tmp_path / "tmp" / "resume_catalog"
    ResumePlan(tmp_path=resume_tmp, progress_bar=False)

    # We'll create mock partial histograms of size 0 and 2
    histogram = SparseHistogram([], [], 0)
    wrong_histogram = SparseHistogram([], [], 2)

    for file_index in range(0, 5):
        ResumePlan.touch_key_done_file(resume_tmp, ResumePlan.SPLITTING_STAGE, f"split_{file_index}")

        histogram_file = ResumePlan.partial_histogram_file(
            tmp_path=resume_tmp, mapping_key=f"map_{file_index}"
        )
        if file_index == 2:
            histogram.to_file(histogram_file)
        else:
            wrong_histogram.to_file(histogram_file)

    with pytest.raises(ValueError, match="histogram partials have incompatible sizes"):
        args = ImportArguments(
            output_artifact_name="resume_catalog",
            input_path=small_sky_parts_dir,
            file_reader="csv",
            output_path=tmp_path,
            dask_tmp=tmp_path,
            tmp_dir=tmp_path,
            resume_tmp=tmp_path / "tmp",
            constant_healpix_order=1,
            pixel_threshold=1000,
            progress_bar=False,
        )
        runner.run(args, dask_client)


@pytest.mark.dask
def test_dask_runner(
    dask_client,
    small_sky_parts_dir,
    assert_parquet_file_ids,
    tmp_path,
):
    """Test basic execution and the types of the written data."""
    args = ImportArguments(
        output_artifact_name="small_sky_object_catalog",
        input_path=small_sky_parts_dir,
        file_reader=CsvReader(
            type_map={
                "ra": np.float32,
                "dec": np.float32,
                "ra_error": np.float32,
                "dec_error": np.float32,
            }
        ),
        output_path=tmp_path,
        dask_tmp=tmp_path,
        highest_healpix_order=1,
        progress_bar=False,
    )

    runner.run(args, dask_client)

    # Check that the catalog metadata file exists
    catalog = read_hats(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert catalog.catalog_info.ra_column == "ra"
    assert catalog.catalog_info.dec_column == "dec"
    assert catalog.catalog_info.total_rows == 131
    assert len(catalog.get_healpix_pixels()) == 1

    # Check that we keep track of the package versions
    assert (
        catalog.catalog_info.hats_builder == f"hats-import v{version('hats-import')}, hats v{version('hats')}"
    )

    # Check that the catalog parquet file exists and contains correct object IDs
    output_file = os.path.join(args.catalog_path, "dataset", "Norder=0", "Dir=0", "Npix=11.parquet")

    expected_ids = [*range(700, 831)]
    assert_parquet_file_ids(output_file, "id", expected_ids)

    # Check that the schema is correct for leaf parquet and _metadata files
    expected_parquet_schema = pa.schema(
        [
            pa.field("_healpix_29", pa.int64()),
            pa.field("id", pa.int64()),
            pa.field("ra", pa.float32()),
            pa.field("dec", pa.float32()),
            pa.field("ra_error", pa.float32()),
            pa.field("dec_error", pa.float32()),
        ]
    )
    metadata = pq.read_metadata(output_file)
    assert metadata.num_row_groups == 1
    assert metadata.row_group(0).column(0).compression == "ZSTD"

    schema = metadata.schema.to_arrow_schema()
    assert schema.equals(expected_parquet_schema)
    schema = pq.read_metadata(args.catalog_path / "dataset" / "_metadata").schema.to_arrow_schema()
    assert schema.equals(expected_parquet_schema)

    # Check that, when re-loaded as a pandas dataframe, the appropriate numeric types are used.
    data_frame = pd.read_parquet(output_file, engine="pyarrow")
    expected_dtypes = pd.Series(
        {
            "_healpix_29": np.int64,
            "id": np.int64,
            "ra": np.float32,
            "dec": np.float32,
            "ra_error": np.float32,
            "dec_error": np.float32,
        }
    )
    assert data_frame.dtypes.equals(expected_dtypes)
    assert data_frame.index.dtype == np.int64

    # Check that the data thumbnail exists
    data_thumbnail_pointer = args.catalog_path / "dataset" / "data_thumbnail.parquet"
    assert data_thumbnail_pointer.exists()
    thumbnail = ParquetFile(data_thumbnail_pointer)
    thumbnail_schema = thumbnail.metadata.schema.to_arrow_schema()
    assert thumbnail_schema.equals(expected_parquet_schema)
    # The thumbnail has 1 row because the catalog has only 1 pixel
    assert len(thumbnail.read()) == 1


@pytest.mark.dask
def test_dask_runner_stats_only(dask_client, small_sky_parts_dir, tmp_path):
    """Test basic execution, without generating catalog parquet outputs."""
    args = ImportArguments(
        output_artifact_name="small_sky",
        input_path=small_sky_parts_dir,
        file_reader="csv",
        output_path=tmp_path,
        dask_tmp=tmp_path,
        highest_healpix_order=1,
        progress_bar=False,
        debug_stats_only=True,
    )

    runner.run(args, dask_client)

    metadata_filename = os.path.join(args.catalog_path, "properties")
    assert os.path.exists(metadata_filename)

    # Check that the catalog parquet file DOES NOT exist
    output_file = os.path.join(args.catalog_path, "dataset", "Norder=0", "Dir=0", "Npix=11.parquet")

    assert not os.path.exists(output_file)

    catalog = read_hats(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert catalog.catalog_info.ra_column == "ra"
    assert catalog.catalog_info.dec_column == "dec"
    assert len(catalog.get_healpix_pixels()) == 1


@pytest.mark.dask
def test_import_mismatch_expectation(
    dask_client,
    small_sky_parts_dir,
    tmp_path,
):
    """Test that the pipeline execution fails if the number of rows does not match
    explicit (but wrong) expectation."""
    args = ImportArguments(
        output_artifact_name="small_sky",
        input_path=small_sky_parts_dir,
        file_reader="csv",
        output_path=tmp_path,
        dask_tmp=tmp_path,
        highest_healpix_order=1,
        progress_bar=False,
        expected_total_rows=1_000,
    )

    with pytest.raises(ValueError, match="does not match expectation"):
        runner.run(args, dask_client)



import pytest
from hats import read_hats

from hats_import.collection.arguments import CollectionArguments
from hats_import.collection.run_import import run


def test_empty_args():
    """Runner should fail with empty arguments"""
    with pytest.raises(TypeError, match="CollectionArguments"):
        run(None, None)


def test_bad_args():
    """Runner should fail with mis-typed arguments"""
    args = {"output_artifact_name": "bad_arg_type"}
    with pytest.raises(TypeError, match="CollectionArguments"):
        run(args, None)


@pytest.mark.dask(timeout=150)
def test_import_collection(
    dask_client,
    small_sky_source_dir,
    tmp_path,
):
    args = (
        CollectionArguments(
            output_artifact_name="small_sky",
            output_path=tmp_path,
            progress_bar=False,
        )
        .catalog(
            input_path=small_sky_source_dir,
            file_reader="csv",
            catalog_type="source",
            ra_column="source_ra",
            dec_column="source_dec",
            sort_columns="source_id",
            highest_healpix_order=2,
        )
        .add_margin(margin_threshold=5.0)
        .add_margin(margin_threshold=50.0)
        .add_index(indexing_column="object_id", include_healpix_29=False)
        .add_index(indexing_column="source_id", include_healpix_29=False)
    )
    run(args, dask_client)

    collection = read_hats(tmp_path / "small_sky")
    assert collection.collection_path == args.catalog_path

    assert collection.all_margins == ["small_sky_5arcs", "small_sky_50arcs"]
    assert collection.all_indexes == {"object_id": "small_sky_object_id", "source_id": "small_sky_source_id"}

    catalog = read_hats(tmp_path / "small_sky" / "small_sky")
    assert catalog.on_disk
    assert catalog.catalog_info.ra_column == "source_ra"
    assert catalog.catalog_info.dec_column == "source_dec"

    catalog = read_hats(tmp_path / "small_sky" / "small_sky_5arcs")
    assert catalog.on_disk
    assert len(catalog.get_healpix_pixels()) == 2
    assert len(catalog) == 4

    catalog = read_hats(tmp_path / "small_sky" / "small_sky_50arcs")
    assert catalog.on_disk
    assert len(catalog.get_healpix_pixels()) == 2
    assert len(catalog) == 17

    catalog = read_hats(tmp_path / "small_sky" / "small_sky_object_id")
    ## This isn't 131, because one object's sources spans two pixels.
    assert catalog.catalog_info.total_rows == 132

    catalog = read_hats(tmp_path / "small_sky" / "small_sky_source_id")
    assert catalog.catalog_info.total_rows == 17161



"""Import a set of non-hats files using dask for parallelization"""

import pickle

import cloudpickle
import hats.pixel_math.healpix_shim as hp
import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
from hats import pixel_math
from hats.io import file_io, paths
from hats.pixel_math.healpix_pixel import HealpixPixel
from hats.pixel_math.sparse_histogram import HistogramAggregator, SparseHistogram
from hats.pixel_math.spatial_index import SPATIAL_INDEX_COLUMN, spatial_index_to_healpix
from upath import UPath

from hats_import.catalog.resume_plan import ResumePlan
from hats_import.pipeline_resume_plan import get_pixel_cache_directory, print_task_failure

# pylint: disable=too-many-locals,too-many-arguments


def _has_named_index(dataframe):
    """Heuristic to determine if a dataframe has some meaningful index.

    This will reject dataframes with no index name for a single index,
    or empty names for multi-index (e.g. [] or [None]).
    """
    if dataframe.index.name is not None:
        ## Single index with a given name.
        return True
    if len(dataframe.index.names) == 0 or all(name is None for name in dataframe.index.names):
        return False
    return True


def _iterate_input_file(
    input_file: UPath,
    pickled_reader_file: str,
    highest_order,
    ra_column,
    dec_column,
    use_healpix_29=False,
    read_columns=None,
):
    """Helper function to handle input file reading and healpix pixel calculation"""
    with open(pickled_reader_file, "rb") as pickle_file:
        file_reader = cloudpickle.load(pickle_file)
    if not file_reader:
        raise NotImplementedError("No file reader implemented")

    for chunk_number, data in enumerate(file_reader.read(input_file, read_columns=read_columns)):
        if use_healpix_29:
            if isinstance(data, pd.DataFrame) and data.index.name == SPATIAL_INDEX_COLUMN:
                mapped_pixels = spatial_index_to_healpix(data.index, target_order=highest_order)
            else:
                mapped_pixels = spatial_index_to_healpix(
                    data[SPATIAL_INDEX_COLUMN], target_order=highest_order
                )
        else:
            # Set up the pixel data
            if isinstance(data, pd.DataFrame):
                mapped_pixels = hp.radec2pix(
                    highest_order,
                    data[ra_column].to_numpy(copy=False, dtype=float),
                    data[dec_column].to_numpy(copy=False, dtype=float),
                )
            else:
                mapped_pixels = hp.radec2pix(
                    highest_order,
                    data[ra_column].to_numpy(),
                    data[dec_column].to_numpy(),
                )
        yield chunk_number, data, mapped_pixels


def map_to_pixels(
    input_file: UPath,
    pickled_reader_file: str,
    resume_path: UPath,
    mapping_key,
    highest_order,
    ra_column,
    dec_column,
    use_healpix_29=False,
):
    """Map a file of input objects to their healpix pixels.

    Args:
        input_file (UPath): file to read for catalog data.
        file_reader (hats_import.catalog.file_readers.InputReader): instance of input
            reader that specifies arguments necessary for reading from the input file.
        resume_path (UPath): where to write resume partial results.
        mapping_key (str): unique counter for this input file, used
            when creating intermediate files
        highest_order (int): healpix order to use when mapping
        ra_column (str): where to find right ascension data in the dataframe
        dec_column (str): where to find declation in the dataframe

    Returns:
        one-dimensional numpy array of long integers where the value at each index corresponds
        to the number of objects found at the healpix pixel.
    Raises:
        ValueError: if the `ra_column` or `dec_column` cannot be found in the input file.
        FileNotFoundError: if the file does not exist, or is a directory
    """
    try:
        histo = HistogramAggregator(highest_order)

        if use_healpix_29:
            read_columns = [SPATIAL_INDEX_COLUMN]
        else:
            read_columns = [ra_column, dec_column]

        for _, _, mapped_pixels in _iterate_input_file(
            input_file,
            pickled_reader_file,
            highest_order,
            ra_column,
            dec_column,
            use_healpix_29,
            read_columns,
        ):
            mapped_pixel, count_at_pixel = np.unique(mapped_pixels, return_counts=True)

            histo.add(SparseHistogram(mapped_pixel, count_at_pixel, highest_order))

        histo.to_sparse().to_file(
            ResumePlan.partial_histogram_file(tmp_path=resume_path, mapping_key=mapping_key)
        )
    except Exception as exception:  # pylint: disable=broad-exception-caught
        print_task_failure(f"Failed MAPPING stage with file {input_file}", exception)
        raise exception


def split_pixels(
    input_file: UPath,
    pickled_reader_file: str,
    splitting_key,
    highest_order,
    ra_column,
    dec_column,
    cache_shard_path: UPath,
    resume_path: UPath,
    alignment_file=None,
    use_healpix_29=False,
):
    """Map a file of input objects to their healpix pixels and split into shards.

    Args:
        input_file (UPath): file to read for catalog data.
        file_reader (hats_import.catalog.file_readers.InputReader): instance
            of input reader that specifies arguments necessary for reading from the input file.
        splitting_key (str): unique counter for this input file, used
            when creating intermediate files
        highest_order (int): healpix order to use when mapping
        ra_column (str): where to find right ascension data in the dataframe
        dec_column (str): where to find declation in the dataframe
        cache_shard_path (UPath): where to write intermediate parquet files.
        resume_path (UPath): where to write resume files.

    Raises:
        ValueError: if the `ra_column` or `dec_column` cannot be found in the input file.
        FileNotFoundError: if the file does not exist, or is a directory
    """
    try:
        with open(alignment_file, "rb") as pickle_file:
            alignment = pickle.load(pickle_file)
        for chunk_number, data, mapped_pixels in _iterate_input_file(
            input_file, pickled_reader_file, highest_order, ra_column, dec_column, use_healpix_29
        ):
            aligned_pixels = alignment[mapped_pixels]
            unique_pixels, unique_inverse = np.unique(aligned_pixels, return_inverse=True, axis=0)

            for unique_index, [order, pixel, _] in enumerate(unique_pixels):
                pixel_dir = get_pixel_cache_directory(cache_shard_path, HealpixPixel(order, pixel))
                file_io.make_directory(pixel_dir, exist_ok=True)
                output_file = file_io.append_paths_to_pointer(
                    pixel_dir, f"shard_{splitting_key}_{chunk_number}.parquet"
                )
                if isinstance(data, pd.DataFrame):
                    filtered_data = data.iloc[unique_inverse == unique_index]
                    if _has_named_index(filtered_data):
                        filtered_data = filtered_data.reset_index()
                    filtered_data = pa.Table.from_pandas(
                        filtered_data, preserve_index=False
                    ).replace_schema_metadata()
                else:
                    filtered_data = data.filter(unique_inverse == unique_index)

                pq.write_table(filtered_data, output_file.path, filesystem=output_file.fs)
                del filtered_data

        ResumePlan.splitting_key_done(tmp_path=resume_path, splitting_key=splitting_key)
    except Exception as exception:  # pylint: disable=broad-exception-caught
        print_task_failure(f"Failed SPLITTING stage with file {input_file}", exception)
        raise exception


# pylint: disable=too-many-positional-arguments
def reduce_pixel_shards(
    cache_shard_path,
    resume_path,
    reducing_key,
    destination_pixel_order,
    destination_pixel_number,
    destination_pixel_size,
    output_path,
    ra_column,
    dec_column,
    sort_columns: str = "",
    use_healpix_29=False,
    add_healpix_29=True,
    delete_input_files=True,
    use_schema_file="",
    write_table_kwargs=None,
    row_group_kwargs=None,
    npix_suffix=".parquet",
):
    """Reduce sharded source pixels into destination pixels.

    In addition to combining multiple shards of data into a single
    parquet file, this method will (optionally) add the ``_healpix_29`` column.
    See ``hats.pixel_math.spatial_index`` for more in-depth discussion of this field.

    Args:
        cache_shard_path (UPath): where to read intermediate parquet files.
        resume_path (UPath): where to write resume files.
        reducing_key (str): unique string for this task, used for resume files.
        destination_pixel_order (int): order of the final catalog pixel
        destination_pixel_number (int): pixel number at the above order
        destination_pixel_size (int): expected number of rows to write
            for the catalog's final pixel
        output_path (UPath): where to write the final catalog pixel data
        ra_column (str): where to find right ascension data in the dataframe
        dec_column (str): where to find declination in the dataframe
        sort_columns (str): column for survey identifier, or other sortable column
        use_healpix_29 (bool): should we use a pre-existing _healpix_29 column
            for position information.
        add_healpix_29 (bool): should we add a _healpix_29 column to
            the resulting parquet file?
        delete_input_files (bool): should we delete the intermediate files
            used as input for this method.
        use_schema_file (str): use the parquet schema from the indicated
            parquet file.
        write_table_kwargs (dict): additional keyword arguments to use when
            writing files to parquet (e.g. compression schemes)
        row_group_kwargs (dict): additional keyword arguments to use in
            creation of rowgroups when writing files to parquet.
        npix_suffix (str): suffix for Npix files. Defaults to ".parquet".

    Raises:
        ValueError: if the number of rows written doesn't equal provided
            `destination_pixel_size`
    """
    try:
        destination_dir = paths.pixel_directory(
            output_path, destination_pixel_order, destination_pixel_number
        )
        file_io.make_directory(destination_dir, exist_ok=True)

        healpix_pixel = HealpixPixel(destination_pixel_order, destination_pixel_number)
        destination_file = paths.pixel_catalog_file(output_path, healpix_pixel, npix_suffix=npix_suffix)
        if destination_file.exists():
            if delete_input_files:
                pixel_dir = get_pixel_cache_directory(cache_shard_path, healpix_pixel)
                file_io.remove_directory(pixel_dir, ignore_errors=True)

            ResumePlan.reducing_key_done(tmp_path=resume_path, reducing_key=reducing_key)
            return

        schema = None
        if use_schema_file:
            schema = file_io.read_parquet_metadata(use_schema_file).schema.to_arrow_schema()

        healpix_pixel = HealpixPixel(destination_pixel_order, destination_pixel_number)
        pixel_dir = get_pixel_cache_directory(cache_shard_path, healpix_pixel)

        merged_table = pq.read_table(pixel_dir, schema=schema)

        rows_written = len(merged_table)

        if rows_written != destination_pixel_size:
            raise ValueError(
                "Unexpected number of objects at pixel "
                f"({healpix_pixel})."
                f" Expected {destination_pixel_size}, wrote {rows_written}"
            )

        if add_healpix_29:
            merged_table = merged_table.add_column(
                0,
                SPATIAL_INDEX_COLUMN,
                [
                    pixel_math.compute_spatial_index(
                        merged_table[ra_column].to_numpy().astype(np.float64),
                        merged_table[dec_column].to_numpy().astype(np.float64),
                    )
                ],
            )

        if not write_table_kwargs:
            write_table_kwargs = {}
        if not row_group_kwargs:
            row_group_kwargs = {}

        sorting_columns = sort_columns.split(",") if sort_columns is not None else []
        if add_healpix_29 or use_healpix_29:
            # Sort by healpix_29 first and then by the other sorting columns to resolve unambiguity
            sorting_columns.insert(0, SPATIAL_INDEX_COLUMN)
        if sorting_columns:
            ordering = [(col_name, "ascending") for col_name in sorting_columns]
            merged_table = merged_table.sort_by(ordering)
            # For metadata purposes this needs to be of type pq.SortingColumn
            write_table_kwargs["sorting_columns"] = pq.SortingColumn.from_ordering(
                merged_table.schema, ordering
            )

        # Obtain the row groups for the target file
        rowgroup_tables = _split_to_row_groups(merged_table, row_group_kwargs, destination_pixel_order)

        with pq.ParquetWriter(
            destination_file.path,
            merged_table.schema,
            filesystem=destination_file.fs,
            **write_table_kwargs,
        ) as writer:
            for table in rowgroup_tables:
                writer.write_table(table)
        del merged_table, rowgroup_tables

        if delete_input_files:
            pixel_dir = get_pixel_cache_directory(cache_shard_path, healpix_pixel)
            file_io.remove_directory(pixel_dir, ignore_errors=True)

        ResumePlan.reducing_key_done(tmp_path=resume_path, reducing_key=reducing_key)
    except Exception as exception:  # pylint: disable=broad-exception-caught
        print_task_failure(
            f"Failed REDUCING stage for shard: {destination_pixel_order} {destination_pixel_number}",
            exception,
        )
        raise exception


def _split_to_row_groups(table, row_group_kwargs, pixel_order):
    """Split the pixel table into its row group chunks according to the specified splitting strategy."""
    if "num_rows" in row_group_kwargs:
        chunk_size = row_group_kwargs["num_rows"]
        return [table.slice(i, chunk_size) for i in range(0, len(table), chunk_size)]
    if "subtile_order_delta" in row_group_kwargs:
        split_tables = []
        parent_pixels = table[SPATIAL_INDEX_COLUMN].to_numpy()
        target_order = row_group_kwargs["subtile_order_delta"] + pixel_order
        child_pixs = spatial_index_to_healpix(parent_pixels, target_order=target_order)
        for child_pix in np.unique(child_pixs):
            indices = np.where(child_pixs == child_pix)[0]
            row_group = table.take(pa.array(indices))
            split_tables.append(row_group)
        return split_tables
    return [table]



"""Tests of argument validation"""

import pytest

from hats_import.verification.arguments import VerificationArguments


def test_invalid_paths(tmp_path, small_sky_object_catalog):
    """Required arguments are provided, but paths aren't found."""
    ## Prove that it works with required args
    VerificationArguments(input_catalog_path=small_sky_object_catalog, output_path=tmp_path)

    # Truth schema is not an existing file
    with pytest.raises(FileNotFoundError, match="truth_schema must be an existing file or directory"):
        VerificationArguments(
            input_catalog_path=small_sky_object_catalog, output_path=tmp_path, truth_schema="path"
        )


@pytest.mark.timeout(5)
def test_good_paths(tmp_path, small_sky_object_catalog):
    """Required arguments are provided, and paths are found.

    NB: This is currently the last test in alpha-order, and may require additional
    time to teardown fixtures."""
    tmp_path_str = str(tmp_path)
    args = VerificationArguments(input_catalog_path=small_sky_object_catalog, output_path=tmp_path)
    assert args.input_catalog_path == small_sky_object_catalog
    assert str(args.output_path) == tmp_path_str
